#!/usr/bin/env python
# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
#
# Produce a valid client.authentication.k8s.io/v1beta1 ExecCredential from
# environment variables.

import json
import os

print(
    json.dumps(
        {
            "apiVersion": "client.authentication.k8s.io/v1beta1",
            "kind": "ExecCredential",
            "status": {
                "clientCertificateData": os.environ["KUBE_CLIENT_CERTIFICATE_DATA"],
                "clientKeyData": os.environ["KUBE_CLIENT_KEY_DATA"],
            },
        }
    )
)



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
import yaml
from jsonpath import JSONPointerKeyError

from kr8s._config import KubeConfig, KubeConfigSet


@pytest.fixture
def temp_kubeconfig(k8s_cluster):
    """Create a temporary kubeconfig by copying the one from k8s_cluster."""
    with open(k8s_cluster.kubeconfig_path, "rb") as f:
        kubeconfig = f.read()
    with NamedTemporaryFile() as f:
        f.write(kubeconfig)
        f.flush()
        yield f.name


async def test_load_kubeconfig(temp_kubeconfig):
    config = await KubeConfig(temp_kubeconfig)
    assert str(config.path) == temp_kubeconfig
    assert config._raw
    assert "current-context" in config._raw
    assert config.current_context
    assert len(config.clusters) > 0
    assert "name" in config.clusters[0]
    assert len(config.users) > 0
    assert len(config.contexts) > 0


async def test_load_kubeconfig_set(temp_kubeconfig):
    configs = await KubeConfigSet(temp_kubeconfig)
    assert len(configs._configs) == 1
    assert str(configs._configs[0].path) == temp_kubeconfig
    assert configs._configs[0]._raw
    assert "current-context" in configs._configs[0]._raw
    assert configs.current_context
    assert len(configs.clusters) > 0
    assert "name" in configs.clusters[0]
    assert len(configs.users) > 0
    assert len(configs.contexts) > 0


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_kubeconfig_from_dict(temp_kubeconfig, cls):
    with open(temp_kubeconfig) as fh:
        config = yaml.safe_load(fh)
    kubeconfig = await cls(config)
    assert kubeconfig.raw == config


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_rename_context(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    context = config.current_context
    new_context = f"{context}-new"
    assert context in [c["name"] for c in config.contexts]
    await config.rename_context(context, new_context)
    # Load again from disk to ensure the change was saved
    config = await cls(temp_kubeconfig)
    assert new_context in [c["name"] for c in config.contexts]
    assert context not in [c["name"] for c in config.contexts]
    assert config.current_context == new_context
    await config.rename_context(new_context, context)
    assert new_context not in [c["name"] for c in config.contexts]
    assert context in [c["name"] for c in config.contexts]
    assert config.current_context == context
    with pytest.raises(ValueError):
        await config.rename_context("foobar", new_context)


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_get_context(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    context = config.current_context
    assert "cluster" in config.get_context(context)
    with pytest.raises(ValueError):
        config.get_context("foobar")
    with pytest.raises(ValueError):
        config.get_context("")


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_get_cluster(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    context = config.get_context(config.current_context)
    assert "server" in config.get_cluster(context["cluster"])
    with pytest.raises(ValueError):
        config.get_cluster("foobar")
    with pytest.raises(ValueError):
        config.get_cluster("")


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_get_user(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    context = config.get_context(config.current_context)
    assert config.get_user(context["user"])
    with pytest.raises(ValueError):
        config.get_user("foobar")
    with pytest.raises(ValueError):
        config.get_user("")


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_use_namespace(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    current_namespace = config.current_namespace
    await config.use_namespace("default")
    assert config.current_namespace == "default"
    await config.use_namespace("kube-system")
    assert config.current_namespace == "kube-system"
    await config.use_namespace(current_namespace)


async def test_kubeconfigset_multi_paths(temp_kubeconfig, tmp_path):
    kubeconfig1 = Path(temp_kubeconfig)
    kubeconfig2 = Path(tmp_path / "kubeconfig")
    kubeconfig2.write_bytes(kubeconfig1.read_bytes())
    config = await KubeConfigSet(kubeconfig1, kubeconfig2)
    context = config.get_context(config.current_context)
    assert config.get_user(context["user"])


async def test_kubeconfigset_multi_paths_one_empty(temp_kubeconfig):
    config = await KubeConfigSet(temp_kubeconfig, "")
    context = config.get_context(config.current_context)
    assert config.get_user(context["user"])


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_kubeconfig_path_empty_fail(cls):
    with pytest.raises(ValueError):
        await cls("")


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_kubeconfig_path_isdir_fail(cls, tmp_path):
    with pytest.raises(IsADirectoryError):
        await cls(tmp_path)


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_get(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    assert isinstance(config.get(path="contexts"), list)
    assert isinstance(config.get(pointer="/contexts/0/name"), str)
    assert isinstance(config.get(path="contexts[*].name"), list)


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_set(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    first_context = config.get(pointer="/contexts/0/name")

    await config.set(pointer="/contexts/0/name", value="foo")
    assert config.get(pointer="/contexts/0/name") == "foo"
    assert config.get(path="contexts[0].name")[0] == "foo"

    await config.set(pointer="/contexts/0/name", value=first_context)
    assert config.get(pointer="/contexts/0/name") == first_context


@pytest.mark.parametrize("cls", [KubeConfig, KubeConfigSet])
async def test_unset(temp_kubeconfig, cls):
    config = await cls(temp_kubeconfig)
    with pytest.raises(JSONPointerKeyError):
        config.get(pointer="/contexts/0/context/foo")
    await config.set(pointer="/contexts/0/context/foo", value="bar")
    assert config.get(pointer="/contexts/0/context/foo") == "bar"
    await config.unset(pointer="/contexts/0/context/foo")
    with pytest.raises(JSONPointerKeyError):
        config.get(pointer="/contexts/0/context/foo")



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import copy
import datetime
import inspect
import pathlib
import platform
import tempfile
import time
from contextlib import suppress

import anyio
import httpx
import pytest

import kr8s
from kr8s._exec import CompletedExec, ExecError
from kr8s.asyncio.objects import (
    APIObject,
    Deployment,
    Ingress,
    PersistentVolume,
    Pod,
    Service,
    get_class,
    new_class,
    object_from_name_type,
    object_from_spec,
    objects_from_files,
)
from kr8s.asyncio.portforward import PortForward
from kr8s.objects import Pod as SyncPod
from kr8s.objects import Service as SyncService
from kr8s.objects import (
    get_class as sync_get_class,
)
from kr8s.objects import (
    new_class as sync_new_class,
)
from kr8s.objects import (
    object_from_spec as sync_object_from_spec,
)
from kr8s.objects import objects_from_files as sync_objects_from_files

DEFAULT_TIMEOUT = httpx.Timeout(30)
CURRENT_DIR = pathlib.Path(__file__).parent


@pytest.fixture
async def nginx_pod(k8s_cluster, example_pod_spec):
    example_pod_spec["metadata"]["name"] = (
        "nginx-" + example_pod_spec["metadata"]["name"]
    )
    example_pod_spec["spec"]["containers"][0]["image"] = "nginx:latest"
    example_pod_spec["spec"]["containers"][0]["ports"] = [{"containerPort": 80}]
    example_pod_spec["spec"]["containers"][0]["readinessProbe"] = {
        "httpGet": {"path": "/", "port": 80},
        "initialDelaySeconds": 0,
        "periodSeconds": 1,
        "timeoutSeconds": 1,
        "successThreshold": 2,
    }
    example_pod_spec["metadata"]["labels"]["app"] = example_pod_spec["metadata"]["name"]
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.ready():
        await anyio.sleep(0.1)
    await pod.exec(
        [
            "dd",
            "if=/dev/random",
            "of=/usr/share/nginx/html/foo.dat",
            "bs=4M",
            "count=10",
        ]
    )
    yield pod
    try:
        await pod.delete()
    except kr8s.NotFoundError:
        pass


@pytest.fixture
async def ubuntu_pod(k8s_cluster, example_pod_spec, ns):
    example_pod_spec["spec"]["containers"][0]["name"] = "ubuntu"
    example_pod_spec["spec"]["containers"][0]["image"] = "ubuntu:latest"
    example_pod_spec["spec"]["containers"][0]["command"] = ["sleep", "3600"]
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.ready():
        await anyio.sleep(0.1)
    yield pod
    await pod.delete()


@pytest.fixture
async def nginx_service(example_service_spec, nginx_pod):
    example_service_spec["metadata"]["name"] = nginx_pod.name
    example_service_spec["spec"]["selector"] = nginx_pod.labels
    service = await Service(example_service_spec)
    await service.create()
    while not await service.ready():
        await anyio.sleep(0.1)  # pragma: no cover
    yield service
    try:
        await service.delete()
    except kr8s.NotFoundError:
        pass


async def test_pod_create_and_delete(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    with pytest.raises(NotImplementedError):
        pod.replicas
    assert await pod.exists()
    while not await pod.ready():
        await anyio.sleep(0.1)
    await pod.delete()
    while await pod.exists():
        await anyio.sleep(0.1)
    assert not await pod.exists()


async def test_pod_object_from_name_type(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert await pod.exists()
    pod2 = await object_from_name_type(f"pod/{pod.name}", namespace=pod.namespace)
    pod3 = await object_from_name_type(f"pod.v1/{pod.name}", namespace=pod.namespace)
    assert pod2.name == pod.name
    assert type(pod2) is type(pod)
    assert pod3.name == pod.name
    assert type(pod3) is type(pod)
    await pod.delete()


async def test_pod_wait_ready(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.wait("condition=Ready")
    with pytest.raises(TimeoutError):
        await pod.wait("jsonpath='{.status.phase}'=Foo", timeout=0.1)
    await pod.wait("condition=Ready=true")
    await pod.wait("condition=Ready=True")
    await pod.wait("jsonpath='{.status.phase}'=Running")
    with pytest.raises(ValueError):
        await pod.wait("foo=NotARealCondition")
    await pod.delete()
    await pod.wait("condition=Ready=False")
    await pod.wait("delete")


async def test_pod_missing_await_error(example_pod_spec):
    pod = Pod(example_pod_spec)  # We intentionally forget to await here
    assert pod._api is None
    with pytest.raises(RuntimeError, match="forget to await it"):
        await pod.create()


async def test_pod_wait_multiple_conditions(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.wait(conditions=["condition=Failed", "condition=Ready"])
    with pytest.raises(TimeoutError):
        await pod.wait(
            conditions=["condition=Failed", "condition=Ready"], mode="all", timeout=0.1
        )
    await pod.wait(
        conditions=[
            "condition=Initialized",
            "condition=ContainersReady",
        ],
        mode="all",
    )
    with pytest.raises(ValueError):
        await pod.wait(conditions=["condition=Failed", "condition=Ready"], mode="foo")
    await pod.delete()


def test_pod_wait_ready_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    pod.wait("condition=Ready")
    with pytest.raises(TimeoutError):
        pod.wait("jsonpath='{.status.phase}'=Foo", timeout=0.1)
    pod.wait("condition=Ready=true")
    pod.wait("condition=Ready=True")
    pod.wait("jsonpath='{.status.phase}'=Running")
    with pytest.raises(ValueError):
        pod.wait("foo=NotARealCondition")
    pod.delete()
    pod.wait("condition=Ready=False")
    pod.wait("delete")


def test_wait_replicas(ns):
    from kr8s.objects import StatefulSet

    ss = StatefulSet(
        {
            "metadata": {
                "name": "test-wait-replicas",
                "namespace": ns,
            },
            "spec": {
                "replicas": 3,
                "selector": {
                    "matchLabels": {
                        "app": "test-wait-replicas",
                    }
                },
                "template": {
                    "metadata": {
                        "labels": {
                            "app": "test-wait-replicas",
                        }
                    },
                    "spec": {
                        "containers": [
                            {
                                "name": "test-wait-replicas",
                                "image": "nginx:latest",
                            }
                        ]
                    },
                },
            },
        }
    )
    ss.create()
    try:
        ss.wait("jsonpath='{.status.availableReplicas}'=3", timeout=30)
    finally:
        ss.delete()


def test_pod_refresh_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    pod.refresh()
    pod.delete()


def test_pod_create_and_delete_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    with pytest.raises(NotImplementedError):
        pod.replicas
    assert pod.exists()
    while not pod.ready():
        time.sleep(0.1)
    pod.delete()
    while pod.exists():
        time.sleep(0.1)
    assert not pod.exists()


async def test_list_and_ensure():
    api = await kr8s.asyncio.api()
    pods = await api.get("pods", namespace=kr8s.ALL)
    assert len(pods) > 0
    for pod in pods:
        await pod.refresh()
        assert await pod.exists(ensure=True)


async def test_nonexistant():
    pod = await Pod(
        {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {
                "name": "nonexistant",
                "namespace": "nonexistant",
            },
        }
    )
    assert not await pod.exists()
    with pytest.raises(kr8s.NotFoundError):
        await pod.exists(ensure=True)


async def test_pod_kind_api_raw():
    pod = await Pod(
        {
            "metadata": {"name": "foo"},
            "spec": {"containers": [{"name": "foo", "image": "nginx"}]},
        }
    )
    assert "kind" in pod.raw
    assert "apiVersion" in pod.raw
    assert pod.raw["kind"] == "Pod"
    assert pod.raw["apiVersion"] == "v1"


async def test_pod_metadata(example_pod_spec, ns):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert "name" in pod.metadata
    assert "hello" in pod.labels
    assert "foo" in pod.annotations
    assert ns == pod.namespace
    assert "example-" in pod.name
    assert "containers" in pod.spec
    assert "phase" in pod.status
    await pod.delete()


async def test_pod_missing_labels_annotations(example_pod_spec):
    del example_pod_spec["metadata"]["labels"]
    del example_pod_spec["metadata"]["annotations"]
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert not pod.labels
    assert not pod.annotations
    await pod.delete()


async def test_pod_get(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    with pytest.raises(kr8s.NotFoundError):
        await Pod.get(f"{pod.name}-foo", namespace=pod.namespace, timeout=0.1)
    pod2 = await Pod.get(pod.name, namespace=pod.namespace)
    assert pod2.name == pod.name
    assert pod2.namespace == pod.namespace
    await pod.delete()
    while await pod.exists():
        await anyio.sleep(0.1)
    with pytest.raises(kr8s.NotFoundError):
        await pod2.delete()


def test_pod_get_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    with pytest.raises(kr8s.NotFoundError):
        SyncPod.get(f"{pod.name}-foo", namespace=pod.namespace, timeout=0.1)
    pod2 = SyncPod.get(pod.name, namespace=pod.namespace)
    assert pod2.name == pod.name
    assert pod2.namespace == pod.namespace
    pod.delete()
    while pod.exists():
        time.sleep(0.1)
    with pytest.raises(kr8s.NotFoundError):
        pod2.delete()


async def test_pod_from_name(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    pod2 = await Pod(
        pod.name, namespace=pod.namespace
    )  # Note: Does not call the Kubernetes API
    assert pod2.name == pod.name
    assert pod2.namespace == pod.namespace
    await pod.delete()
    while await pod.exists():
        await anyio.sleep(0.1)
    with pytest.raises(kr8s.NotFoundError):
        await pod2.delete()


async def test_pod_get_timeout(example_pod_spec):
    async def create_pod():
        await anyio.sleep(0.1)
        pod = await Pod(example_pod_spec)
        await pod.create()
        return pod

    async def get_pod():
        pod = await Pod.get(
            example_pod_spec["metadata"]["name"],
            namespace=example_pod_spec["metadata"]["namespace"],
            timeout=1,
        )
        return pod

    async with anyio.create_task_group() as tg:
        tg.start_soon(create_pod)
        tg.start_soon(get_pod)
    pod = await get_pod()
    await pod.delete()


async def test_missing_pod():
    with pytest.raises(kr8s.NotFoundError):
        await Pod.get("nonexistant", namespace="default")


@pytest.mark.parametrize("selector", ["abc=123def", {"abc": "123def"}])
async def test_label_selector(example_pod_spec, selector):
    example_pod_spec["metadata"]["labels"]["abc"] = "123def"
    pod = await Pod(example_pod_spec)
    await pod.create()

    api = await kr8s.asyncio.api()
    pods = await api.get("pods", namespace=kr8s.ALL, label_selector=selector)
    assert len(pods) >= 0

    await pod.delete()


async def test_field_selector(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()

    api = await kr8s.asyncio.api()
    pods = await api.get(
        "pods", namespace=kr8s.ALL, field_selector={"metadata.name": pod.name}
    )
    assert len(pods) == 1

    pods = await api.get(
        "pods", namespace=kr8s.ALL, field_selector="metadata.name=" + "foo-bar-baz"
    )
    assert len(pods) == 0

    await pod.delete()


async def test_get_with_label_selector(example_pod_spec, ns):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.label(test="test_get_with_label_selector")

    pod2 = await Pod.get(label_selector=pod.labels, namespace=ns)
    assert pod == pod2

    pod3 = await Pod.get(field_selector={"metadata.name": pod.name}, namespace=ns)
    assert pod == pod3

    await pod.delete()


async def test_pod_watch(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    async for event, obj in pod.watch():
        assert event in ("ADDED", "MODIFIED", "DELETED")
        assert obj.name == pod.name
        break
    await pod.delete()


async def test_pod_annotate(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.annotate({"foo": "bar"})
    await pod.annotate(fizz="buzz")
    assert "foo" in pod.annotations
    assert "fizz" in pod.annotations
    with pytest.raises(ValueError):
        await pod.annotate({})
    await pod.delete()


async def test_pod_label(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    await pod.label({"foo": "bar"})
    assert "foo" in pod.labels
    with pytest.raises(ValueError):
        await pod.label({})
    await pod.delete()


def test_pod_watch_sync(example_pod_spec):
    pod = SyncPod(example_pod_spec)
    pod.create()
    for event, obj in pod.watch():
        assert event in ("ADDED", "MODIFIED", "DELETED")
        assert obj.name == pod.name
        break
    pod.delete()


async def test_patch_pod(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert "patched" not in pod.labels
    await pod.patch({"metadata": {"labels": {"patched": "true"}}})
    assert "patched" in pod.labels
    await pod.delete()


async def test_patch_pod_json(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert "patched" not in pod.labels
    await pod.patch(
        [{"op": "replace", "path": "/metadata/labels", "value": {"patched": "true"}}],
        type="json",
    )
    assert set(pod.labels) == {"patched"}
    await pod.delete()


async def test_all_v1_objects_represented():
    api = await kr8s.asyncio.api()
    k8s_objects = await api.api_resources()
    supported_apis = (
        "v1",
        "apps/v1",
        "autoscaling/v2",
        "batch/v1",
        "networking.k8s.io/v1",
        "policy/v1",
        "rbac.authorization.k8s.io/v1",
        "apiextensions.k8s.io/v1",
    )
    # for supported_api in supported_apis:
    #     assert supported_api in [obj["version"] for obj in objects]
    objects = [obj for obj in k8s_objects if obj["version"] in supported_apis]
    for obj in objects:
        assert get_class(obj["kind"], obj["version"])


async def test_object_from_spec(example_pod_spec, example_service_spec):
    pod = object_from_spec(example_pod_spec)
    assert isinstance(pod, Pod)
    assert pod.name == example_pod_spec["metadata"]["name"]
    assert pod.spec == example_pod_spec["spec"]
    assert pod._asyncio

    service = object_from_spec(example_service_spec)
    assert isinstance(service, Service)
    assert service.name == example_service_spec["metadata"]["name"]
    assert service.spec == example_service_spec["spec"]
    assert service._asyncio


async def test_object_from_spec_sync(example_pod_spec, example_service_spec):
    pod = sync_object_from_spec(example_pod_spec)
    assert isinstance(pod, Pod)
    assert pod.name == example_pod_spec["metadata"]["name"]
    assert pod.spec == example_pod_spec["spec"]
    assert not pod._asyncio

    service = sync_object_from_spec(example_service_spec)
    assert isinstance(service, Service)
    assert service.name == example_service_spec["metadata"]["name"]
    assert service.spec == example_service_spec["spec"]
    assert not service._asyncio


async def test_subclass_registration():
    with pytest.raises(KeyError):
        get_class("MyResource", "foo.kr8s.org/v1alpha1")

    class MyResource(APIObject):
        version = "foo.kr8s.org/v1alpha1"
        endpoint = "myresources"
        kind = "MyResource"
        plural = "myresources"
        singular = "myresource"
        namespaced = True

    get_class("MyResource", "foo.kr8s.org/v1alpha1")


async def test_new_class_registration():
    with pytest.raises(KeyError):
        get_class("MyOtherResource", "foo.kr8s.org/v1alpha1")

    MyOtherResource = new_class("MyOtherResource.foo.kr8s.org/v1alpha1")  # noqa: F841

    get_class("MyOtherResource", "foo.kr8s.org/v1alpha1")
    assert MyOtherResource._asyncio


async def test_new_sync_class_registration():
    with pytest.raises(KeyError):
        sync_get_class("MyOtherSyncResource", "foo.kr8s.org/v1alpha1")

    MyOtherSyncResource = sync_new_class(
        "MyOtherSyncResource.foo.kr8s.org/v1alpha1"
    )  # noqa: F841

    sync_get_class("MyOtherSyncResource", "foo.kr8s.org/v1alpha1")
    assert not MyOtherSyncResource._asyncio


async def test_new_class_registration_from_spec():
    my_async_resource_instance = await object_from_spec(
        {
            "kind": "MyAsyncResource",
            "apiVersion": "foo.kr8s.org/v1alpha1",
            "metadata": {"name": "foo"},
            "spec": {},
        },
        allow_unknown_type=True,
    )  # noqa: F841

    assert my_async_resource_instance._asyncio


async def test_new_sync_class_registration_from_spec():
    my_sync_resource_instance = sync_object_from_spec(
        {
            "kind": "MySyncResource",
            "apiVersion": "foo.kr8s.org/v1alpha1",
            "metadata": {"name": "foo"},
            "spec": {},
        },
        allow_unknown_type=True,
    )  # noqa: F841

    assert not my_sync_resource_instance._asyncio


async def test_class_registration_multiple_subclass():
    class MyResource(new_class("MyResource.foo.kr8s.org/v1alpha1")):
        def my_custom_method(self) -> str:
            return "foo"

    assert get_class("MyResource", "foo.kr8s.org/v1alpha1") is MyResource

    r = MyResource({})
    assert r.my_custom_method() == "foo"


async def test_deployment_scale(example_deployment_spec):
    deployment = await Deployment(example_deployment_spec)
    await deployment.create()
    assert deployment.replicas == 1
    await deployment.scale(2)
    assert deployment.replicas == 2
    while not await deployment.ready():
        await anyio.sleep(0.1)
    pods = await deployment.pods()
    assert len(pods) == 2
    await deployment.scale(1)
    assert deployment.replicas == 1
    await deployment.delete()


async def test_node():
    api = await kr8s.asyncio.api()
    nodes = await api.get("nodes")
    assert len(nodes) > 0
    for node in nodes:
        assert node.unschedulable is False
        await node.cordon()
        assert node.unschedulable is True
        await node.uncordon()


async def test_service_proxy():
    api = await kr8s.asyncio.api()
    [service] = await api.get("services", "kubernetes")
    assert service.name == "kubernetes"
    data = await service.proxy_http_get("/version", raise_for_status=False)
    assert isinstance(data, httpx.Response)


async def test_pod_logs(example_pod_spec):
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.ready():
        await anyio.sleep(0.1)
    log = "\n".join([line async for line in pod.logs(container="pause")])
    assert isinstance(log, str)
    await pod.delete()


async def test_pod_port_forward_context_manager(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    async with nginx_pod.portforward(80) as port:
        async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
            resp = await session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = await session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404
            resp = await session.get(f"http://localhost:{port}/foo.dat")
            assert resp.status_code == 200
            resp.read()


def test_pod_port_forward_context_manager_sync(nginx_service):
    nginx_service = SyncService.get(
        nginx_service.name, namespace=nginx_service.namespace
    )
    with nginx_service.portforward(80) as port:
        with httpx.Client(timeout=DEFAULT_TIMEOUT) as session:
            resp = session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404
            resp = session.get(f"http://localhost:{port}/foo.dat")
            assert resp.status_code == 200
            resp.read()


@pytest.mark.skip(reason="For manual testing only")
async def test_pod_port_forward_context_manager_manual(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    pf = nginx_pod.portforward(80, 8184)
    async with pf:
        done = False
        while not done:
            # Put a breakpoint here and set done = True when you're finished.
            await anyio.sleep(1)


async def test_pod_port_forward_start_stop(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    for _ in range(5):
        pf = nginx_pod.portforward(80)
        assert pf._bg_task is None
        port = await pf.start()
        assert pf._bg_task is not None
        async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
            resp = await session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = await session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404
            resp = await session.get(f"http://localhost:{port}/foo.dat")
            assert resp.status_code == 200
            resp.read()
        await pf.stop()
        assert pf._bg_task is None


async def test_service_port_forward_context_manager(nginx_service):
    async with nginx_service.portforward(80) as port:
        async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
            resp = await session.get(f"http://localhost:{port}/")
            assert resp.status_code == 200
            resp = await session.get(f"http://localhost:{port}/foo")
            assert resp.status_code == 404


async def test_service_port_forward_start_stop(nginx_service):
    pf = nginx_service.portforward(80)
    assert pf._bg_task is None
    port = await pf.start()
    assert pf._bg_task is not None

    async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
        resp = await session.get(f"http://localhost:{port}/")
        assert resp.status_code == 200
        resp = await session.get(f"http://localhost:{port}/foo")
        assert resp.status_code == 404

    await pf.stop()
    assert pf._bg_task is None


async def test_unsupported_port_forward():
    pv = await PersistentVolume({"metadata": {"name": "foo"}})
    with pytest.raises(AttributeError):
        await pv.portforward(80)
    with pytest.raises(ValueError):
        await PortForward(pv, 80).start()


@pytest.mark.skipif(
    "macOS" in platform.platform(),
    reason="Hangs on macOS, see https://github.com/kr8s-org/kr8s/issues/380",
)
async def test_multiple_bind_addresses_port_forward(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()

    # Example multiple addresses
    multiple_addresses = ["127.0.0.2", "127.0.0.3"]

    pf = nginx_pod.portforward(80, local_port=None, address=multiple_addresses)

    # Start the port forwarding
    await pf.start()
    async with httpx.AsyncClient(timeout=DEFAULT_TIMEOUT) as session:
        for address in multiple_addresses:
            resp = await session.get(f"http://{address}:{pf.local_port}/")
            assert resp.status_code == 200

    # Stop the port forwarding
    await pf.stop()


async def test_scalable_dot_notation():
    class Foo(APIObject):
        version = "foo.kr8s.org/v1alpha1"
        endpoint = "foos"
        kind = "Foo"
        plural = "foos"
        singular = "foo"
        namespaced = True
        scalable = True
        scalable_spec = "nested.replicas"

    foo = await Foo({"metadata": {"name": "foo"}, "spec": {"nested": {"replicas": 1}}})
    assert foo.replicas == 1


async def test_object_from_file():
    api = await kr8s.asyncio.api()
    objects = await objects_from_files(
        CURRENT_DIR / "resources" / "simple" / "nginx_pod.yaml", api=api
    )
    assert len(objects) == 1
    assert isinstance(objects[0], Pod)
    assert objects[0].kind == "Pod"
    assert objects[0].name == "nginx"
    assert len(objects[0].spec.containers) == 1


async def test_objects_from_file():
    objects = await objects_from_files(
        CURRENT_DIR / "resources" / "simple" / "nginx_pod_service.yaml"
    )
    assert len(objects) == 2
    assert isinstance(objects[0], Pod)
    assert isinstance(objects[1], Service)
    assert len(objects[1].spec.ports) == 1


def test_objects_from_file_sync():
    objects = sync_objects_from_files(
        CURRENT_DIR / "resources" / "simple" / "nginx_pod_service.yaml"
    )
    assert len(objects) == 2
    assert isinstance(objects[0], Pod)
    assert isinstance(objects[1], Service)
    assert not objects[0]._asyncio
    assert not objects[0].api._asyncio
    assert len(objects[1].spec.ports) == 1


async def test_objects_from_files():
    simple_dir = CURRENT_DIR / "resources" / "simple"
    objects = await objects_from_files(simple_dir)
    assert len(objects) > 1


async def test_objects_from_files_nested():
    simple_dir = CURRENT_DIR / "resources" / "simple"

    objects = await objects_from_files(simple_dir)
    assert not any(isinstance(o, Ingress) for o in objects)

    objects = await objects_from_files(simple_dir, recursive=True)
    assert any(isinstance(o, Ingress) for o in objects)


async def test_custom_object_from_file():
    simple_dir = CURRENT_DIR / "resources" / "custom" / "evc.yaml"
    objects = await objects_from_files(simple_dir)
    assert len(objects) == 1


async def test_pod_to_dict(example_pod_spec):
    pod = Pod(example_pod_spec)
    assert dict(pod) == example_pod_spec
    assert dict(pod) == pod.raw


async def test_adoption(nginx_service):
    [nginx_pod, *_] = await nginx_service.ready_pods()
    await nginx_service.adopt(nginx_pod)
    assert "ownerReferences" in nginx_pod.metadata
    assert nginx_pod.metadata["ownerReferences"][0]["name"] == nginx_service.name
    await nginx_service.delete()
    while await nginx_pod.exists():
        await anyio.sleep(0.1)


async def test_cast_to_from_lightkube(example_pod_spec):
    pytest.importorskip("lightkube")
    from lightkube import codecs
    from lightkube.resources.core_v1 import Pod as LightkubePod

    starting_pod = codecs.from_dict(example_pod_spec)

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"

    lightkube_pod = kr8s_pod.to_lightkube()
    assert isinstance(lightkube_pod, LightkubePod)
    assert lightkube_pod.metadata.name == example_pod_spec["metadata"]["name"]
    assert lightkube_pod.metadata.namespace == example_pod_spec["metadata"]["namespace"]


async def test_cast_to_from_kubernetes(example_pod_spec):
    kubernetes = pytest.importorskip("kubernetes")

    starting_pod = kubernetes.client.models.v1_pod.V1Pod(
        api_version=example_pod_spec["apiVersion"],
        kind=example_pod_spec["kind"],
        metadata=example_pod_spec["metadata"],
        spec=example_pod_spec["spec"],
    )

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"


async def test_cast_to_from_kubernetes_asyncio(example_pod_spec):
    kubernetes_asyncio = pytest.importorskip("kubernetes_asyncio")

    starting_pod = kubernetes_asyncio.client.models.v1_pod.V1Pod(
        api_version=example_pod_spec["apiVersion"],
        kind=example_pod_spec["kind"],
        metadata=example_pod_spec["metadata"],
        spec=example_pod_spec["spec"],
    )

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"


async def test_cast_to_from_pykube_ng(example_pod_spec):
    pykube = pytest.importorskip("pykube")

    starting_pod = pykube.objects.Pod(None, example_pod_spec)

    kr8s_pod = await Pod(starting_pod)
    assert isinstance(kr8s_pod, Pod)
    assert kr8s_pod.name == example_pod_spec["metadata"]["name"]
    assert kr8s_pod.namespace == example_pod_spec["metadata"]["namespace"]
    assert kr8s_pod.kind == "Pod"
    assert kr8s_pod.version == "v1"

    pykube_pod = kr8s_pod.to_pykube(None)
    assert isinstance(pykube_pod, pykube.objects.Pod)
    assert pykube_pod.name == example_pod_spec["metadata"]["name"]
    assert pykube_pod.namespace == example_pod_spec["metadata"]["namespace"]


async def test_to_dict(example_pod_spec):
    pod = await Pod(example_pod_spec)
    to_spec = pod.to_dict()
    assert to_spec == example_pod_spec
    assert isinstance(to_spec, dict)


async def test_pod_exec(ubuntu_pod):
    ex = await ubuntu_pod.exec(["date"])
    assert isinstance(ex, CompletedExec)
    assert str(datetime.datetime.now().year) in ex.stdout.decode()
    assert ex.args == ["date"]
    assert ex.stderr == b""
    assert ex.returncode == 0


async def test_pod_exec_error(ubuntu_pod):
    with pytest.raises(ExecError):
        await ubuntu_pod.exec(["date", "foo"])

    ex = await ubuntu_pod.exec(["date", "foo"], check=False)
    assert ex.args == ["date", "foo"]
    assert b"invalid date" in ex.stderr
    assert ex.returncode == 1

    with pytest.raises(ExecError):
        ex.check_returncode()


async def test_pod_exec_to_file(ubuntu_pod):
    with tempfile.TemporaryFile(mode="w+b") as tmp:
        exc = await ubuntu_pod.exec(["date"], stdout=tmp, capture_output=False)
        tmp.seek(0)
        assert str(datetime.datetime.now().year) in tmp.read().decode()
        assert exc.stdout == b""

    with tempfile.TemporaryFile(mode="w+b") as tmp:
        with pytest.raises(ExecError):
            await ubuntu_pod.exec(["date", "foo"], stderr=tmp)
        tmp.seek(0)
        assert b"invalid date" in tmp.read()


@pytest.mark.xfail(reason="Exec protocol v5.channel.k8s.io not available")
async def test_pod_exec_stdin(ubuntu_pod):
    ex = await ubuntu_pod.exec(["cat"], stdin="foo")
    assert b"foo" in ex.stdout


async def test_pod_exec_not_ready(ns):
    pod = await Pod.gen(name="nginx", namespace=ns, image="nginx:latest")
    await pod.create()
    try:
        assert not await pod.ready()
        await pod.exec(["date"])
        assert await pod.ready()
    finally:
        await pod.delete()


async def test_configmap_data(ns):
    [cm] = await objects_from_files(CURRENT_DIR / "resources" / "configmap.yaml")
    cm.namespace = ns
    await cm.create()
    assert "game.properties" in cm.data
    assert cm.data.player_initial_lives == "3"
    assert "color.good=purple" in cm.data["user-interface.properties"]
    await cm.delete()


async def test_secret_data(ns):
    [secret] = await objects_from_files(CURRENT_DIR / "resources" / "secret.yaml")
    secret.namespace = ns
    await secret.create()
    assert "tls.crt" in secret.data
    await secret.delete()


async def test_validate_pod(example_pod_spec):
    kubernetes_validate = pytest.importorskip("kubernetes_validate")
    pod = await Pod(example_pod_spec)
    kubernetes_validate.validate(pod.raw, "1.28", strict=True)


async def test_validate_pod_fail(bad_pod_spec):
    kubernetes_validate = pytest.importorskip("kubernetes_validate")
    pod = await Pod(bad_pod_spec)
    with pytest.raises(kubernetes_validate.ValidationError):
        kubernetes_validate.validate(pod.raw, "1.28", strict=True)


async def test_pod_errors(bad_pod_spec):
    pod = await Pod(bad_pod_spec)
    with pytest.raises(kr8s.ServerError, match="Required value"):
        await pod.create()


async def test_pod_list():
    pods1 = await kr8s.asyncio.get("pods", namespace=kr8s.ALL)
    pods2 = await Pod.list(namespace=kr8s.ALL)
    assert pods1 and pods2
    assert len(pods1) == len(pods2)
    assert all(isinstance(p, Pod) for p in pods1)
    assert all(isinstance(p, Pod) for p in pods2)
    assert {p.name for p in pods1} == {p.name for p in pods2}


@pytest.mark.parametrize(
    "ports",
    [
        80,
        [80],
        [80, 81],
        [{"containerPort": 80}],
        [{"containerPort": 80}, {"containerPort": 81}],
    ],
)
async def test_pod_gen_ports(ns, ports):
    pod = await Pod.gen(name="nginx", namespace=ns, image="nginx:latest", ports=ports)
    try:
        await pod.create()  # This should succeed
    finally:
        with suppress(kr8s.NotFoundError):
            await pod.delete()


def test_sync_new_class_is_sync():
    MyResource = new_class(
        kind="MyResource",
        version="newclass.example.com/v1",
        namespaced=True,
        asyncio=False,
    )
    instance = MyResource({})
    assert not instance._asyncio
    assert not inspect.iscoroutinefunction(instance.create)


def test_new_class_plural_suffix():
    MyFoo = new_class(
        kind="MyFoo",
        version="newclass.example.com/v1",
        namespaced=True,
    )
    instance = MyFoo({})
    assert instance.plural == "myfoos"
    assert instance.endpoint == "myfoos"

    MyClass = new_class(
        kind="MyClass",
        plural="MyClasses",
        version="newclass.example.com/v1",
        namespaced=True,
    )
    instance = MyClass({})
    assert instance.plural == "myclasses"
    assert instance.endpoint == "myclasses"

    MyPolicy = new_class(
        kind="MyPolicy",
        plural="MyPolicies",
        version="newclass.example.com/v1",
        namespaced=True,
    )
    instance = MyPolicy({})
    assert instance.plural == "mypolicies"
    assert instance.endpoint == "mypolicies"


def test_object_setter(example_pod_spec):
    po = Pod(example_pod_spec)

    assert po.name != "foo"
    po.raw["metadata"]["name"] = "foo"
    assert po.name == "foo"

    assert po.raw["spec"]["containers"][0]["name"] != "bar"
    po.raw["spec"]["containers"][0]["name"] = "bar"
    assert po.raw["spec"]["containers"][0]["name"] == "bar"


def test_object_setter_from_old_spec(example_pod_spec):
    spec = copy.deepcopy(example_pod_spec)

    po = Pod(example_pod_spec)

    assert po.raw["spec"]["containers"][0]["name"] != "bar"
    po.raw["spec"]["containers"][0]["name"] = "bar"
    assert po.raw["spec"]["containers"][0]["name"] == "bar"

    new_po = Pod(spec)
    assert new_po.raw["spec"]["containers"][0]["name"] != "bar"
    new_po.raw["spec"] = po.raw["spec"]
    assert new_po.raw["spec"]["containers"][0]["name"] == "bar"


def test_parse_kind():
    from kr8s._objects import parse_kind

    assert parse_kind("Pod") == ("pod", "", "")
    assert parse_kind("Pods") == ("pods", "", "")
    assert parse_kind("pod/v1") == ("pod", "", "v1")
    assert parse_kind("deploy") == ("deploy", "", "")
    assert parse_kind("gateway") == ("gateway", "", "")
    assert parse_kind("gateways") == ("gateways", "", "")
    assert parse_kind("gateway.networking.istio.io") == (
        "gateway",
        "networking.istio.io",
        "",
    )
    assert parse_kind("gateways.networking.istio.io") == (
        "gateways",
        "networking.istio.io",
        "",
    )
    assert parse_kind("gateway.v1.networking.istio.io") == (
        "gateway",
        "networking.istio.io",
        "v1",
    )
    assert parse_kind("gateways.v1.networking.istio.io") == (
        "gateways",
        "networking.istio.io",
        "v1",
    )
    assert parse_kind("gateway.networking.istio.io/v1") == (
        "gateway",
        "networking.istio.io",
        "v1",
    )
    assert parse_kind("gateways.networking.istio.io/v1") == (
        "gateways",
        "networking.istio.io",
        "v1",
    )


async def test_setting_attributes():
    po = await Pod.gen(name="nginx", image="nginx:latest")
    po.metadata.labels = {"foo": "bar"}
    assert po.metadata.labels == {"foo": "bar"}

    po.metadata.generateName = po.metadata.pop("name") + "-"
    assert "generateName" in po.metadata
    assert "name" not in po.metadata

    po.name = "abc123"
    assert po.name == "abc123"
    po.namespace = "bar"
    assert po.namespace == "bar"
    po.metadata = {"name": "def", "namespace": "buzz"}
    assert po.name == "def"
    po["metadata"] = {"name": "ghi", "namespace": "buzz"}
    assert po.name == "ghi"

    po.spec.containers[0].image = "wordpress:latest"
    assert po.spec.containers[0].image == "wordpress:latest"

    with pytest.raises(NotImplementedError):
        po.replicas = 2


async def test_generate_name():
    po = await kr8s.asyncio.objects.Pod.gen(
        generate_name="nginx-", image="nginx:latest"
    )

    assert "generateName" in po.metadata
    assert po.metadata.generateName == "nginx-"
    assert "name" not in po.metadata
    with pytest.raises(ValueError):
        assert po.name

    assert "generateName(nginx-)" in po.__repr__()

    await po.create()
    try:
        assert po.name
        assert po.name.startswith("nginx-")
        assert len(po.name) > len(po.metadata.generateName)
        assert po.metadata.generateName in po.name
    finally:
        await po.delete()



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import datetime

project = "kr8s"
author = "Dask Developers, NVIDIA"
copyright = f"{datetime.date.today().year}, {author}"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.intersphinx",
    "sphinx_design",
    "sphinx_copybutton",
    "sphinxcontrib.mermaid",
    "myst_parser",
    "sphinx.ext.autodoc",
    "autoapi.extension",
]
suppress_warnings = ["autoapi"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- copybutton configuration ---------------------------------------------------

copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- autoapi configuration ---------------------------------------------------

autodoc_typehints = "signature"  # autoapi respects this

autoapi_type = "python"
autoapi_dirs = ["../kr8s"]
autoapi_template_dir = "_templates/autoapi"
autoapi_options = [
    "members",
    "inherited-members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "imported-members",
]
autoapi_ignore = ["*tests*", "*conftest*"]
# autoapi_python_use_implicit_namespaces = True
autoapi_keep_files = True
# autoapi_generate_api_docs = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]

html_theme_options = {
    "light_logo": "branding/logo-solo.png",
    "dark_logo": "branding/logo-solo.png",
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}


def remove_async_property(app, what, name, obj, skip, options):
    """Remove async property from sync wrapped methods.

    Find all sync classes that inherit from async classes
    and remove the async property from their wrapped methods.

    Also skip public ``async_`` methods as they are only intended
    to be used internally by other sync wrapped objects.
    """
    if what == "class" and (
        # Infer that the class is a sync wrapped class if it is the kr8s.Api class,
        # or is in kr8s.objects and inherits from a class in kr8s._objects.
        name == "kr8s.Api"
        or ("kr8s.objects" in name and obj.bases and "kr8s._objects" in obj.bases[0])
        # FIXME: It would be better to just check if the class is decorated with @sync
        # but sphinx-autoapi does not currently keep track of decorators
        # See https://github.com/readthedocs/sphinx-autoapi/issues/459
    ):
        for child in obj.children:
            if (
                child.type == "method"
                and not child.name.startswith("async_")
                and not child.name.startswith("_")
                and "async" in child.properties
            ):
                child.properties.remove("async")

    if (
        what == "method"
        and ("kr8s.objects" in name or "kr8s.asyncio.objects" in name)
        and obj.name.startswith("async_")
    ):
        skip = True
    return skip


def setup(app):
    app.add_css_file("css/custom.css")
    app.connect("autoapi-skip-member", remove_async_property)



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from typing import List

import anyio
import rich.table
import typer
from rich import box
from rich.console import Console
from rich.live import Live

import kr8s
from kr8s.asyncio.objects import Table

TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

console = Console()


async def draw_table(kind, response, resource_names):
    table = rich.table.Table(box=box.SIMPLE)
    table.add_column("Namespace", style="magenta", no_wrap=True)

    for column in response.column_definitions:
        if column["priority"] == 0:
            kwargs = {}
            if column["name"] == "Name":
                kwargs = {"style": "cyan", "no_wrap": True}
            table.add_column(column["name"], **kwargs)

    for row in response.rows:
        if not resource_names or row["object"]["metadata"]["name"] == resource_names[0]:
            r = [
                str(row)
                for row, column in zip(row["cells"], response.column_definitions)
                if column["priority"] == 0
            ]
            table.add_row(row["object"]["metadata"]["namespace"], *r)

    if not table.rows:
        if resource_names:
            console.print(f"No resources found with name {resource_names[0]}.")
        else:
            console.print(f"No {kind} resources found in current namespace.")
        return None
    return table


async def get_resources(resources, label_selector, field_selector):
    data = {}
    for kind in resources:
        data[kind] = await kr8s.asyncio.get(
            kind,
            label_selector=label_selector,
            field_selector=field_selector,
            as_object=Table,
        )
    return data


async def get(
    resources: List[str] = typer.Argument(..., help="TYPE[.VERSION][.GROUP]"),
    all_namespaces: bool = typer.Option(
        False,
        "-A",
        "--all-namespaces",
        help="If present, list the requested object(s) across all namespaces. "
        "Namespace in current context is ignored even if specified with --namespace.",
    ),
    namespace: str = typer.Option(
        None,
        "-n",
        "--namespace",
    ),
    label_selector: str = typer.Option(
        "",
        "-l",
        "--selector",
        help="Selector (label query) to filter on, supports '=', '==', and '!='. "
        "(e.g. -l key1=value1,key2=value2). "
        "Matching objects must satisfy all of the specified label constraints.",
    ),
    field_selector: str = typer.Option(
        "",
        "--field-selector",
        help="Selector (field query) to filter on, supports '=', '==', and '!='. "
        "(e.g. --field-selector key1=value1,key2=value2). "
        "The server only supports a limited number of field queries per type. ",
    ),
    show_kind: bool = typer.Option(
        False,
        "--show-kind",
        help="If present, list the resource type for the requested object(s).",
    ),
    show_labels: bool = typer.Option(
        False,
        "--show-labels",
        help="When printing, show all labels as the last column "
        "(default hide labels column).",
    ),
    label_columns: List[str] = typer.Option(
        [],
        "-L",
        "--label-columns",
        help="Accepts a comma separated list of labels that are going to be presented as columns. "
        "Names are case-sensitive. "
        "You can also use multiple flag options like -L label1 -L label2...",
    ),
    watch: bool = typer.Option(
        False,
        "-w",
        "--watch",
        help="After listing/getting the requested object, watch for changes.",
    ),
):
    """Display one or many resources.

    Prints a table of the most important information about the specified resources. You can filter the list using a
    label selector and the --selector flag. If the desired resource type is namespaced you will only see results
    in your current namespace unless you pass --all-namespaces.

    By specifying the output as 'template' and providing a Jinja2 template as the value of the --template flag, you
    can filter the attributes of the fetched resources.

    Use "kubectl-ng api-resources" for a complete list of supported resources.
    """
    resource_names = []
    # Support kubectl-ng get pod,svc
    if len(resources) == 1:
        resources = resources[0].split(",")
        if len(resources) > 1 and watch:
            raise typer.BadParameter("you may only specify a single resource type")
    # Support kubectl-ng get pod [name]
    elif len(resources) == 2:
        resources, resource_names = ([r] for r in resources)
    else:
        raise typer.BadParameter(
            f"Format '{' '.join(resources)}' not currently supported."
        )
    kubernetes = await kr8s.asyncio.api()
    if namespace:
        kubernetes.namespace = namespace
    if all_namespaces:
        kubernetes.namespace = kr8s.ALL

    data = await get_resources(resources, label_selector, field_selector)
    for kind, response in data.items():
        table = await draw_table(kind, response, resource_names)
    if not watch:
        console.print(table)
    else:
        with Live(table, console=console, auto_refresh=False) as live:
            while True:
                await anyio.sleep(5)
                data = await get_resources(resources, label_selector, field_selector)
                # TODO handle changes in resources
                for kind, response in data.items():
                    table = await draw_table(kind, response, resource_names)
                live.update(table)
                live.refresh()



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import queue
import threading

import anyio
import pytest

import kr8s
import kr8s.asyncio
from kr8s._exceptions import APITimeoutError
from kr8s.asyncio.objects import Pod, Table


async def test_factory_bypass() -> None:
    with pytest.raises(ValueError, match="kr8s.api()"):
        _ = kr8s.Api()
    assert not kr8s.Api._instances
    _ = kr8s.api()
    assert kr8s.Api._instances


async def test_api_factory(serviceaccount) -> None:
    k1 = await kr8s.asyncio.api()
    k2 = await kr8s.asyncio.api()
    assert k1 is k2

    k3 = await kr8s.asyncio.api(serviceaccount=serviceaccount)
    k4 = await kr8s.asyncio.api(serviceaccount=serviceaccount)
    assert k1 is not k3
    assert k3 is k4

    p = await Pod({"metadata": {"name": "foo"}})
    assert p.api is k1
    assert p.api is not k3


def test_api_factory_threaded() -> None:
    assert len(kr8s.Api._instances) == 0

    q = queue.Queue()

    def run_in_thread(q):
        async def create_api(q):
            k = await kr8s.asyncio.api()
            q.put(k)

        anyio.run(create_api, q)

    t1 = threading.Thread(
        target=run_in_thread,
        args=(q,),
    )
    t2 = threading.Thread(
        target=run_in_thread,
        args=(q,),
    )
    t1.start()
    t2.start()
    t1.join()
    t2.join()
    k1 = q.get()
    k2 = q.get()

    assert k1 is not k2
    assert type(k1) is type(k2)


def test_api_factory_multi_event_loop() -> None:
    assert len(kr8s.Api._instances) == 0

    async def create_api():
        return await kr8s.asyncio.api()

    k1 = anyio.run(create_api)
    k2 = anyio.run(create_api)
    assert k1 is not k2


async def test_api_factory_with_kubeconfig(k8s_cluster, serviceaccount) -> None:
    k1 = await kr8s.asyncio.api(kubeconfig=k8s_cluster.kubeconfig_path)
    k2 = await kr8s.asyncio.api(serviceaccount=serviceaccount)
    k3 = await kr8s.asyncio.api()
    assert k1 is not k2
    assert k3 is k1
    assert k3 is not k2

    p = await Pod({"metadata": {"name": "foo"}})
    assert p.api is k1

    p2 = await Pod({"metadata": {"name": "bar"}}, api=k2)
    assert p2.api is k2

    p3 = await Pod({"metadata": {"name": "baz"}}, api=k3)
    assert p3.api is k3
    assert p3.api is not k2


def test_version_sync() -> None:
    api = kr8s.api()
    version = api.version()
    assert "major" in version


async def test_version_sync_in_async() -> None:
    api = kr8s.api()
    version = api.version()
    assert "major" in version


async def test_version() -> None:
    api = await kr8s.asyncio.api()
    version = await api.version()
    assert "major" in version


def test_helper_version() -> None:
    version = kr8s.version()
    assert "major" in version


async def test_concurrent_api_creation() -> None:
    async def get_api():
        api = await kr8s.asyncio.api()
        await api.version()

    async with anyio.create_task_group() as tg:
        for _ in range(10):
            tg.start_soon(get_api)


async def test_both_api_creation_methods_together() -> None:
    async_api = await kr8s.asyncio.api()
    api = kr8s.api()

    assert await kr8s.asyncio.api() is async_api
    assert kr8s.api() is api
    assert async_api is not api

    assert await async_api.version() == api.version()
    assert await async_api.whoami() == api.whoami()

    assert (await async_api.get("ns"))[0]._asyncio is True
    assert api.get("ns")[0]._asyncio is False


async def test_bad_api_version() -> None:
    api = await kr8s.asyncio.api()
    with pytest.raises(ValueError):
        async with api.call_api("GET", version="foo"):
            pass  # pragma: no cover


@pytest.mark.parametrize("namespace", [kr8s.ALL, "kube-system"])
async def test_get_pods(namespace) -> None:
    pods = await kr8s.asyncio.get("pods", namespace=namespace)
    assert isinstance(pods, list)
    assert len(pods) > 0
    assert isinstance(pods[0], Pod)


async def test_get_pods_as_table() -> None:
    api = await kr8s.asyncio.api()
    pods = await api.get("pods", namespace="kube-system", as_object=Table)
    assert isinstance(pods, Table)
    assert len(pods.rows) > 0
    assert not await pods.exists()  # Cannot exist in the Kubernetes API


async def test_watch_pods(example_pod_spec, ns) -> None:
    pod = await Pod(example_pod_spec)
    await pod.create()
    while not await pod.ready():
        await anyio.sleep(0.1)
    async for event, obj in kr8s.asyncio.watch("pods", namespace=ns):
        assert event in ["ADDED", "MODIFIED", "DELETED"]
        assert isinstance(obj, Pod)
        if obj.name == pod.name:
            if event == "ADDED":
                await obj.patch({"metadata": {"labels": {"test": "test"}}})
            elif event == "MODIFIED" and "test" in obj.labels and await obj.exists():
                await obj.delete()
                while await obj.exists():
                    await anyio.sleep(0.1)
            elif event == "DELETED":
                break


async def test_get_deployments() -> None:
    api = await kr8s.asyncio.api()
    deployments = await api.get("deployments")
    assert isinstance(deployments, list)


async def test_get_class() -> None:
    api = await kr8s.asyncio.api()
    pods = await api.get(Pod, namespace=kr8s.ALL)
    assert isinstance(pods, list)
    assert len(pods) > 0
    assert isinstance(pods[0], Pod)


async def test_api_versions() -> None:
    api = await kr8s.asyncio.api()
    versions = [version async for version in api.api_versions()]
    assert "apps/v1" in versions


def test_api_versions_sync() -> None:
    api = kr8s.api()
    versions = [version for version in api.api_versions()]
    assert "apps/v1" in versions


async def test_api_resources() -> None:
    resources = await kr8s.asyncio.api_resources()

    names = [r["name"] for r in resources]
    assert "nodes" in names
    assert "pods" in names
    assert "services" in names
    assert "namespaces" in names

    [pods] = [r for r in resources if r["name"] == "pods"]
    assert pods["namespaced"]
    assert pods["kind"] == "Pod"
    assert pods["version"] == "v1"
    assert "get" in pods["verbs"]

    [deployment] = [d for d in resources if d["name"] == "deployments"]
    assert deployment["namespaced"]
    assert deployment["kind"] == "Deployment"
    assert deployment["version"] == "apps/v1"
    assert "get" in deployment["verbs"]
    assert "deploy" in deployment["shortNames"]


async def test_ns(ns) -> None:
    api = await kr8s.asyncio.api(namespace=ns)
    assert ns == api.namespace

    api.namespace = "foo"
    assert api.namespace == "foo"


async def test_async_get_returns_async_objects() -> None:
    pods = await kr8s.asyncio.get("pods", namespace=kr8s.ALL)
    assert pods[0]._asyncio is True


def test_sync_get_returns_sync_objects() -> None:
    pods = kr8s.get("pods", namespace=kr8s.ALL)
    assert pods[0]._asyncio is False


def test_sync_api_returns_sync_objects() -> None:
    api = kr8s.api()
    pods = api.get("pods", namespace=kr8s.ALL)
    assert pods[0]._asyncio is False


async def test_api_names(example_pod_spec: dict, ns: str) -> None:
    pod = await Pod(example_pod_spec)
    await pod.create()
    assert pod in await kr8s.asyncio.get("pods", namespace=ns)
    assert pod in await kr8s.asyncio.get("pods/v1", namespace=ns)
    assert pod in await kr8s.asyncio.get("Pod", namespace=ns)
    assert pod in await kr8s.asyncio.get("pod", namespace=ns)
    assert pod in await kr8s.asyncio.get("po", namespace=ns)
    await pod.delete()

    await kr8s.asyncio.get("roles", namespace=ns)
    await kr8s.asyncio.get("roles.rbac.authorization.k8s.io", namespace=ns)
    await kr8s.asyncio.get("roles.v1.rbac.authorization.k8s.io", namespace=ns)
    await kr8s.asyncio.get("roles.rbac.authorization.k8s.io/v1", namespace=ns)


async def test_whoami() -> None:
    api = await kr8s.asyncio.api()
    assert await kr8s.asyncio.whoami() == await api.whoami()


async def test_whoami_sync() -> None:
    api = kr8s.api()
    assert kr8s.whoami() == api.whoami()


async def test_api_resources_cache(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level("INFO")
    api = await kr8s.asyncio.api()
    await api.api_resources()
    assert caplog.text.count('/apis/ "HTTP/1.1 200 OK"') == 1
    await api.api_resources()
    assert caplog.text.count('/apis/ "HTTP/1.1 200 OK"') == 1


async def test_api_timeout() -> None:
    from httpx import Timeout

    api = await kr8s.asyncio.api()
    api.timeout = 10
    await api.version()
    assert api._session
    assert api._session.timeout.read == 10
    api.timeout = 20
    await api.version()
    assert api._session.timeout.read == 20
    api.timeout = Timeout(30)
    await api.version()
    assert api._session.timeout.read == 30

    api.timeout = 0.00001
    with pytest.raises(APITimeoutError):
        await api.version()


async def test_lookup_kind():
    api = await kr8s.asyncio.api()

    assert await api.lookup_kind("no") == ("node/v1", "nodes", False)
    assert await api.lookup_kind("nodes") == ("node/v1", "nodes", False)
    assert await api.lookup_kind("po") == ("pod/v1", "pods", True)
    assert await api.lookup_kind("pods/v1") == ("pod/v1", "pods", True)
    assert await api.lookup_kind("CSIStorageCapacity") == (
        "csistoragecapacity.storage.k8s.io/v1",
        "csistoragecapacities",
        True,
    )
    assert await api.lookup_kind("role") == (
        "role.rbac.authorization.k8s.io/v1",
        "roles",
        True,
    )
    assert await api.lookup_kind("roles") == (
        "role.rbac.authorization.k8s.io/v1",
        "roles",
        True,
    )
    assert await api.lookup_kind("roles.v1.rbac.authorization.k8s.io") == (
        "role.rbac.authorization.k8s.io/v1",
        "roles",
        True,
    )
    assert await api.lookup_kind("roles.rbac.authorization.k8s.io") == (
        "role.rbac.authorization.k8s.io/v1",
        "roles",
        True,
    )


async def test_nonexisting_resource_type():
    api = await kr8s.asyncio.api()

    with pytest.raises(ValueError):
        await api.get("foo.bar.baz/v1")


@pytest.mark.parametrize(
    "kind",
    [
        "csr",
        "certificatesigningrequest",
        "certificatesigningrequests",
        "certificatesigningrequest.certificates.k8s.io",
        "certificatesigningrequests.certificates.k8s.io",
        "certificatesigningrequest.v1.certificates.k8s.io",
        "certificatesigningrequests.v1.certificates.k8s.io",
        "certificatesigningrequest.certificates.k8s.io/v1",
        "certificatesigningrequests.certificates.k8s.io/v1",
    ],
)
async def test_dynamic_classes(kind, ensure_gc):
    from kr8s.asyncio.objects import get_class

    api = await kr8s.asyncio.api()

    with pytest.raises(KeyError):
        get_class("certificatesigningrequest", "certificates.k8s.io/v1")

    with pytest.raises(KeyError):
        await api.get(kind, allow_unknown_type=False)

    await api.get(kind)


@pytest.mark.parametrize(
    "kind",
    [
        "ingress.networking.k8s.io",
        "networkpolicies.networking.k8s.io",
        "csistoragecapacities.storage.k8s.io",
        "CSIStorageCapacity",
    ],
)
async def test_get_dynamic_plurals(kind, ensure_gc):
    api = await kr8s.asyncio.api()
    assert isinstance(await api.get(kind), list)



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import datetime
import pathlib

import pytest
from kubectl_ng.cli import app
from typer.testing import CliRunner

from kr8s.objects import objects_from_files

runner = CliRunner()

HERE = pathlib.Path(__file__).parent.absolute()


@pytest.fixture
def pod_service():
    spec = str(HERE / "resources" / "simple" / "nginx_pod_service.yaml")
    objs = objects_from_files(spec)
    for obj in objs:
        obj.create()
    yield [obj.name for obj in objs]
    for obj in objs:
        obj.delete()


def test_create_and_delete(pod_service):
    pod_name, _ = pod_service

    result = runner.invoke(app, ["exec", pod_name, "--", "date"])
    assert result.exit_code == 0
    assert str(datetime.datetime.now().year) in result.stdout



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from os import PathLike
from typing import (
    TYPE_CHECKING,
    Iterable,
    List,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

_KT = TypeVar("_KT")
_VT_co = TypeVar("_VT_co", covariant=True)
PathType = Union[
    str,
    "PathLike[str]",  # Can remove quotes when Python 3.9 is the minimum version.
]

if TYPE_CHECKING:
    from ._objects import Pod


@runtime_checkable
class APIObjectWithPods(Protocol):
    """An APIObject subclass that contains other Pod objects."""

    async def async_ready_pods(self) -> List["Pod"]: ...


@runtime_checkable
class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    """Copied from _typeshed.SupportsKeysAndGetItem to avoid importing it and make runtime checkable."""

    def keys(self) -> Iterable[_KT]: ...
    def __getitem__(self, key: _KT, /) -> _VT_co: ...


class SupportsToDict(Protocol):
    """An object that can be converted to a dictionary."""

    def to_dict(self) -> dict: ...


class SupportsObjAttr(Protocol):
    """An object that has an `obj` dict attribute."""

    obj: dict


# Type that can be converted to a Kubernetes spec.
SpecType = Union[dict, str, SupportsKeysAndGetItem, SupportsToDict, SupportsObjAttr]



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import typer

from ._api_resources import api_resources
from ._api_versions import api_versions
from ._config import config
from ._create import create
from ._delete import delete
from ._exec import kexec
from ._get import get
from ._typer_utils import register
from ._version import version
from ._wait import wait

app = typer.Typer(no_args_is_help=True)
register(app, api_resources)
register(app, api_versions)
register(app, create)
register(app, delete)
register(app, get)
register(app, version)
register(app, wait)
register(app, kexec, "exec")
register(app, config, "config")


def go():
    app()


if __name__ == "__main__":
    go()



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License

import typer
from rich.console import Console

import kr8s.asyncio
from kr8s.asyncio.objects import objects_from_files

console = Console()


# Missing Options
# TODO --allow-missing-template-keys='true'
# TODO --dry-run='none'
# TODO --edit='false'
# TODO --field-manager=''
# TODO -k, --kustomize=''
# TODO -o, --output=''
# TODO --raw='false'
# TODO -R, --recursive='false'
# TODO --save-config='false'
# TODO -l, --selector=''
# TODO --show-managed-fields='false'
# TODO --template=''
# TODO --validate='true'
# TODO --windows-line-endings='false'


async def create(
    filename: str = typer.Option(
        "",
        "--filename",
        "-f",
        help="Filename, directory, or URL to files identifying the resources to create",
    ),
):
    api = await kr8s.asyncio.api()
    try:
        objs = await objects_from_files(filename, api)
    except Exception as e:
        console.print(f"[red]Error loading objects from {filename}[/red]: {e}")
        raise typer.Exit(1)
    for obj in objs:
        try:
            await obj.create()
        except Exception as e:
            console.print(f"[red]Error creating {obj}[/red]: {e}")
            raise typer.Exit(1)
        console.print(f'[green]{obj.singular} "{obj}" created [/green]')



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from kubectl_ng.cli import app
from typer.testing import CliRunner

import kr8s

runner = CliRunner()


def test_current_context():
    current_context = kr8s.api().auth.kubeconfig.current_context
    result = runner.invoke(app, ["config", "current-context"])
    assert result.exit_code == 0
    assert current_context in result.stdout


def test_get_clusters(k8s_cluster):
    result = runner.invoke(app, ["config", "get-clusters"])
    assert result.exit_code == 0
    assert k8s_cluster.name in result.stdout


def test_get_users(k8s_cluster):
    result = runner.invoke(app, ["config", "get-users"])
    assert result.exit_code == 0
    assert k8s_cluster.name in result.stdout


def test_get_contexts(k8s_cluster):
    result = runner.invoke(app, ["config", "get-contexts"])
    assert result.exit_code == 0
    assert k8s_cluster.name in result.stdout

    result = runner.invoke(app, ["config", "get-contexts", f"kind-{k8s_cluster.name}"])
    assert result.exit_code == 0
    assert k8s_cluster.name in result.stdout

    result = runner.invoke(app, ["config", "get-contexts", "foo"])
    assert result.exit_code == 1
    assert "foo not found" in result.stdout


def test_use_context():
    current_context = kr8s.api().auth.kubeconfig.current_context
    result = runner.invoke(app, ["config", "use-context", current_context])
    assert result.exit_code == 0
    assert current_context in result.stdout


def test_rename_context():
    # Get current context
    current_context = kr8s.api().auth.kubeconfig.current_context
    result = runner.invoke(app, ["config", "current-context"])
    assert result.exit_code == 0
    assert current_context in result.stdout

    # Rename current context to foo
    result = runner.invoke(app, ["config", "rename-context", current_context, "foo"])
    assert result.exit_code == 0
    assert current_context in result.stdout
    assert "foo" in result.stdout

    # Check the context rename was successful
    result = runner.invoke(app, ["config", "current-context"])
    assert result.exit_code == 0
    assert "foo" in result.stdout

    # Rename foo back to the original name
    result = runner.invoke(app, ["config", "rename-context", "foo", current_context])
    assert result.exit_code == 0
    assert current_context in result.stdout
    assert "foo" in result.stdout

    # Check the context revert was successful
    result = runner.invoke(app, ["config", "current-context"])
    assert result.exit_code == 0
    assert current_context in result.stdout



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from typing import Optional

import rich.table
import typer
from rich import box
from rich.console import Console
from typing_extensions import Annotated

import kr8s

from ._typer_utils import register

console = Console()

config = typer.Typer(
    no_args_is_help=True,
    name="config",
    help="Modify kubeconfig files.",
)


def config_current_context():
    """Display the current context."""
    try:
        typer.echo(kr8s.api().auth.kubeconfig.current_context)
    except KeyError:
        typer.echo("error: current-context is not set")
        raise typer.Exit(code=1)


def config_get_clusters():
    """Display clusters defined in the kubeconfig."""
    try:
        clusters = [cluster["name"] for cluster in kr8s.api().auth.kubeconfig.clusters]
    except:  # noqa
        clusters = []

    table = rich.table.Table(box=box.SIMPLE)
    table.add_column("Name", style="magenta", no_wrap=True)
    for cluster in clusters:
        table.add_row(cluster)

    console.print(table)


def config_get_users():
    """Display users defined in the kubeconfig."""
    try:
        users = [user["name"] for user in kr8s.api().auth.kubeconfig.users]
    except:  # noqa
        users = []

    table = rich.table.Table(box=box.SIMPLE)
    table.add_column("Name", style="magenta", no_wrap=True)
    for user in users:
        table.add_row(user)

    console.print(table)


def config_get_contexts(name: Annotated[Optional[str], typer.Argument()] = None):
    """Display users defined in the kubeconfig."""
    try:
        kubeconfig = kr8s.api().auth.kubeconfig
        contexts = kubeconfig.contexts
    except:  # noqa
        contexts = []

    table = rich.table.Table(box=box.SIMPLE)
    table.add_column("Current", no_wrap=True)
    table.add_column("Name", style="magenta", no_wrap=True)
    table.add_column("Cluster", style="blue", no_wrap=True)
    table.add_column("Auth Info", style="orange3", no_wrap=True)
    table.add_column("Namespace", style="yellow", no_wrap=True)
    for context in contexts:
        if name and context["name"] != name:
            continue
        table.add_row(
            "*" if context["name"] == kubeconfig.current_context else "",
            context["name"],
            context["context"]["cluster"],
            context["context"]["user"],
            context["context"].get("namespace", ""),
        )
    if name and not table.rows:
        typer.echo(f"error: context {name} not found")
        raise typer.Exit(code=1)

    console.print(table)


async def config_use_context(context: Annotated[str, typer.Argument()]):
    """Set the current-context in a kubeconfig file."""
    try:
        api = await kr8s.asyncio.api()
        await api.auth.kubeconfig.use_context(context)
    except (ValueError, KeyError):
        typer.echo(f"error: context {context} not found")
        raise typer.Exit(code=1)

    console.print(f'Switched to context "{context}".')


async def config_rename_context(
    old_name: Annotated[str, typer.Argument()],
    new_name: Annotated[str, typer.Argument()],
):
    """Renames a context from the kubeconfig file."""
    try:
        api = await kr8s.asyncio.api()
        await api.auth.kubeconfig.rename_context(old_name, new_name)
    except (ValueError, KeyError):
        typer.echo(f"error: context {old_name} not found")
        raise typer.Exit(code=1)

    console.print(f'Context "{old_name}" renamed to "{new_name}".')


register(config, config_current_context, "current-context")
register(config, config_get_clusters, "get-clusters")
register(config, config_get_users, "get-users")
register(config, config_get_contexts, "get-contexts")
register(config, config_use_context, "use-context")
register(config, config_rename_context, "rename-context")



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""Objects for managing a port forward connection.

This module provides a class for managing a port forward connection to a Kubernetes Pod or Service.
"""
import threading
import time

from ._async_utils import sync
from ._portforward import PortForward as _PortForward


@sync
class PortForward(_PortForward):
    __doc__ = _PortForward.__doc__
    _bg_thread = None

    def start(self):
        """Start a background thread with the port forward running."""
        self._bg_thread = threading.Thread(target=self.run_forever, daemon=True)
        self._bg_thread.start()

    def stop(self):
        """Stop the background thread."""
        for server in self.servers:
            while server is None:
                time.sleep(0.1)
            server.close()



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import pytest

from kr8s._data_utils import (
    dict_list_pack,
    dict_to_selector,
    dot_to_nested_dict,
    list_dict_unpack,
    xdict,
)


def test_list_dict_unpack():
    data = [{"key": "hello", "value": "world"}]
    assert list_dict_unpack(data) == {"hello": "world"}


def test_dict_list_pack():
    data = {"hello": "world"}
    assert dict_list_pack(data) == [{"key": "hello", "value": "world"}]


def test_pack_unpack():
    data = [{"key": "hello", "value": "world"}]
    assert dict_list_pack(list_dict_unpack(data)) == data

    data = {"hello": "world"}
    assert list_dict_unpack(dict_list_pack(data)) == data


def test_unpack_pack_deduplicate():
    data = [
        {"key": "hello", "value": "world"},
        {"key": "hello", "value": "there"},
    ]
    unpacked = list_dict_unpack(data)
    assert unpacked["hello"] == "there"
    repacked = dict_list_pack(unpacked)
    assert len(repacked) == 1
    assert repacked == [{"key": "hello", "value": "there"}]


def test_dot_to_nested_dict():
    assert dot_to_nested_dict("hello", "world") == {"hello": "world"}
    assert dot_to_nested_dict("hello.world", "value") == {"hello": {"world": "value"}}
    assert dot_to_nested_dict("foo.bar.baz", "value") == {
        "foo": {
            "bar": {"baz": "value"},
        }
    }


def test_dict_to_selector():
    assert dict_to_selector({"foo": "bar"}) == "foo=bar"
    assert dict_to_selector({"foo": "bar", "baz": "qux"}) == "foo=bar,baz=qux"


def test_xdict():
    assert xdict(foo="bar") == {"foo": "bar"}
    assert xdict(foo="bar", baz=None) == {"foo": "bar"}
    assert xdict(foo="bar", baz="qux") == {"foo": "bar", "baz": "qux"}
    assert xdict(foo="bar", baz="qux", quux=None) == {"foo": "bar", "baz": "qux"}
    assert xdict({"foo": "bar", "baz": "qux"}) == {"foo": "bar", "baz": "qux"}
    assert xdict({"foo": "bar", "baz": None}) == {"foo": "bar"}
    with pytest.raises(ValueError):
        xdict({}, {})
    with pytest.raises(ValueError):
        assert xdict({}, quux=None) == {
            "foo": "bar",
            "baz": "qux",
        }



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import gc
import os
import time

import pytest
from pytest_kind.cluster import KindCluster


@pytest.fixture
def ensure_gc():
    """Ensure garbage collection is run before and after the test."""
    gc.collect()
    yield
    gc.collect()


@pytest.fixture(scope="session", autouse=True)
def k8s_cluster(request) -> KindCluster:
    image = None
    if version := os.environ.get("KUBERNETES_VERSION"):
        image = f"kindest/node:v{version}"

    kind_cluster = KindCluster(
        name="pytest-kind",
        image=image,
    )
    kind_cluster.create()
    os.environ["KUBECONFIG"] = str(kind_cluster.kubeconfig_path)
    # CI fix, wait for default service account to be created before continuing
    while True:
        try:
            kind_cluster.kubectl("get", "serviceaccount", "default")
            break
        except Exception:
            time.sleep(1)
    yield kind_cluster
    del os.environ["KUBECONFIG"]
    if not request.config.getoption("keep_cluster"):  # pragma: no cover
        kind_cluster.delete()



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License

import asyncio
from contextlib import suppress
from functools import wraps

import typer


def _typer_async(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        with suppress(asyncio.CancelledError, KeyboardInterrupt):
            return asyncio.run(f(*args, **kwargs))

    return wrapper


def register(app, func, alias=None):
    if asyncio.iscoroutinefunction(func):
        func = _typer_async(func)
    if isinstance(func, typer.Typer):
        assert alias, "Typer subcommand must have an alias."
        app.add_typer(func, name=alias)
    else:
        if alias is not None:
            app.command(alias)(func)
        else:
            app.command()(func)



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from kubectl_ng.cli import app
from typer.testing import CliRunner

runner = CliRunner()


def test_help_default():
    result = runner.invoke(app, [])
    assert result.exit_code == 0
    assert "Usage:" in result.stdout



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License

from typing import Optional

import httpx


class NotFoundError(Exception):
    """Unable to find the requested resource."""


class ConnectionClosedError(Exception):
    """A connection has been closed."""


class APITimeoutError(Exception):
    """A timeout has occurred while waiting for a response from the Kubernetes API server."""


class ExecError(Exception):
    """Internal error in the exec protocol."""


class ServerError(Exception):
    """Error from the Kubernetes API server.

    Attributes:
        status: The Status object from the Kubernetes API server
        response: The httpx response object
    """

    def __init__(
        self,
        message: str,
        status: Optional[str] = None,
        response: Optional[httpx.Response] = None,
    ) -> None:
        self.status = status
        self.response = response
        super().__init__(message)



#!/usr/bin/env python

# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import json
import re
import urllib.request
from datetime import datetime
from pathlib import Path

from ruamel.yaml import YAML

yaml = YAML(typ="rt")
yaml.width = 2**16
yaml.preserve_quotes = True
yaml.indent(mapping=2, sequence=4, offset=2)

DATE_FORMAT = "%Y-%m-%d"


def get_versions():
    print(
        "Loading Kubernetes versions from https://endoflife.date/api/kubernetes.json..."
    )
    with urllib.request.urlopen("https://endoflife.date/api/kubernetes.json") as url:
        data = json.load(url)
        data = [
            {
                "cycle": x["cycle"],
                "latest_version": x["latest"],
                "eol": datetime.strptime(x["eol"], DATE_FORMAT),
            }
            for x in data
            if datetime.strptime(x["eol"], DATE_FORMAT) > datetime.now()
        ]
        data.sort(key=lambda x: x["eol"], reverse=True)

    print("Loading Kubernetes tags from https://hub.docker.com/r/kindest/node/tags...")

    container_tags = []
    next_url = "https://hub.docker.com/v2/repositories/kindest/node/tags"
    while next_url:
        with urllib.request.urlopen(next_url) as url:
            results = json.load(url)
            container_tags += results["results"]
            if "next" in results and results["next"]:
                next_url = results["next"]
            else:
                next_url = None

    for version in data:
        try:
            version["latest_kind_container"] = [
                x["name"]
                for x in container_tags
                if version["cycle"] in x["name"] and "alpha" not in x["name"]
            ][0][1:]
        except IndexError:
            version["latest_kind_container"] = None

    before_length = len(data)
    print("Pruning versions that do not have a kind release yet...")
    data[:] = [x for x in data if x["latest_kind_container"] is not None]
    print(f"Pruned {before_length - len(data)} versions")
    return data


def update_workflow(versions, workflow_path):
    workflow_path = Path(workflow_path)
    workflow = yaml.load(workflow_path)
    workflow["jobs"]["test"]["strategy"]["matrix"]["kubernetes-version"][0] = versions[
        0
    ]["latest_kind_container"]
    workflow["jobs"]["test"]["strategy"]["matrix"]["include"] = []
    for version in versions[1:]:
        workflow["jobs"]["test"]["strategy"]["matrix"]["include"].append(
            {
                "python-version": "3.10",
                "kubernetes-version": version["latest_kind_container"],
            }
        )
    yaml.dump(workflow, workflow_path)


def update_badges(filename, versions):
    readme = Path(filename).read_text()
    # Use regex to replace the badge
    v = [x["cycle"] for x in versions]
    v.sort()
    version_list = "%7C".join(v)
    readme = re.sub(
        r"img.shields.io/badge/Kubernetes%20support.*-blue",
        f"img.shields.io/badge/Kubernetes%20support-{version_list}-blue",
        readme,
    )
    Path(filename).write_text(readme)


def main():
    versions = get_versions()
    print(f"Latest version: {versions[0]['cycle']}")
    print("Supported versions:")
    for version in versions:
        print(
            f"For {version['cycle']} using kindest/node {version['latest_kind_container']} until {version['eol']}"
        )

    update_workflow(versions, ".github/workflows/test-kr8s.yaml")
    update_workflow(versions, ".github/workflows/test-kubectl-ng.yaml")
    update_badges("README.md", versions)
    update_badges("docs/index.md", versions)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""Objects to represent Kubernetes resources.

This module provides classes that represent Kubernetes resources.
These classes are used to interact with resources in the Kubernetes API server.
"""
from functools import partial

from ._async_utils import run_sync, sync
from ._objects import (
    APIObject as _APIObject,
)
from ._objects import (
    Binding as _Binding,
)
from ._objects import (
    ClusterRole as _ClusterRole,
)
from ._objects import (
    ClusterRoleBinding as _ClusterRoleBinding,
)
from ._objects import (
    ComponentStatus as _ComponentStatus,
)
from ._objects import (
    ConfigMap as _ConfigMap,
)
from ._objects import (
    ControllerRevision as _ControllerRevision,
)
from ._objects import (
    CronJob as _CronJob,
)
from ._objects import (
    CustomResourceDefinition as _CustomResourceDefinition,
)
from ._objects import (
    DaemonSet as _DaemonSet,
)
from ._objects import (
    Deployment as _Deployment,
)
from ._objects import (
    Endpoints as _Endpoints,
)
from ._objects import (
    Event as _Event,
)
from ._objects import (
    HorizontalPodAutoscaler as _HorizontalPodAutoscaler,
)
from ._objects import (
    Ingress as _Ingress,
)
from ._objects import (
    IngressClass as _IngressClass,
)
from ._objects import (
    Job as _Job,
)
from ._objects import (
    LimitRange as _LimitRange,
)
from ._objects import (
    Namespace as _Namespace,
)
from ._objects import (
    NetworkPolicy as _NetworkPolicy,
)
from ._objects import (
    Node as _Node,
)
from ._objects import (
    PersistentVolume as _PersistentVolume,
)
from ._objects import (
    PersistentVolumeClaim as _PersistentVolumeClaim,
)
from ._objects import (
    Pod as _Pod,
)
from ._objects import (
    PodDisruptionBudget as _PodDisruptionBudget,
)
from ._objects import (
    PodTemplate as _PodTemplate,
)
from ._objects import (
    ReplicaSet as _ReplicaSet,
)
from ._objects import (
    ReplicationController as _ReplicationController,
)
from ._objects import (
    ResourceQuota as _ResourceQuota,
)
from ._objects import (
    Role as _Role,
)
from ._objects import (
    RoleBinding as _RoleBinding,
)
from ._objects import (
    Secret as _Secret,
)
from ._objects import (
    Service as _Service,
)
from ._objects import (
    ServiceAccount as _ServiceAccount,
)
from ._objects import (
    StatefulSet as _StatefulSet,
)
from ._objects import (
    Table as _Table,
)
from ._objects import (
    get_class as _get_class,
)
from ._objects import (
    new_class as _new_class,
)

# noqa
from ._objects import object_from_name_type as _object_from_name_type
from ._objects import (
    object_from_spec as _object_from_spec,
)
from ._objects import objects_from_files as _objects_from_files


@sync
class APIObject(_APIObject):
    __doc__ = _APIObject.__doc__
    _asyncio = False


@sync
class Binding(_Binding):
    __doc__ = _Binding.__doc__
    _asyncio = False


@sync
class ComponentStatus(_ComponentStatus):
    __doc__ = _ComponentStatus.__doc__
    _asyncio = False


@sync
class ConfigMap(_ConfigMap):
    __doc__ = _ConfigMap.__doc__
    _asyncio = False


@sync
class Endpoints(_Endpoints):
    __doc__ = _Endpoints.__doc__
    _asyncio = False


@sync
class Event(_Event):
    __doc__ = _Event.__doc__
    _asyncio = False


@sync
class LimitRange(_LimitRange):
    __doc__ = _LimitRange.__doc__
    _asyncio = False


@sync
class Namespace(_Namespace):
    __doc__ = _Namespace.__doc__
    _asyncio = False


@sync
class Node(_Node):
    __doc__ = _Node.__doc__
    _asyncio = False


@sync
class PersistentVolume(_PersistentVolume):
    __doc__ = _PersistentVolume.__doc__
    _asyncio = False


@sync
class PersistentVolumeClaim(_PersistentVolumeClaim):
    __doc__ = _PersistentVolumeClaim.__doc__
    _asyncio = False


@sync
class Pod(_Pod):
    __doc__ = _Pod.__doc__
    _asyncio = False


@sync
class PodTemplate(_PodTemplate):
    __doc__ = _PodTemplate.__doc__
    _asyncio = False


@sync
class ReplicationController(_ReplicationController):
    __doc__ = _ReplicationController.__doc__
    _asyncio = False


@sync
class ResourceQuota(_ResourceQuota):
    __doc__ = _ResourceQuota.__doc__
    _asyncio = False


@sync
class Secret(_Secret):
    __doc__ = _Secret.__doc__
    _asyncio = False


@sync
class Service(_Service):
    __doc__ = _Service.__doc__
    _asyncio = False


@sync
class ServiceAccount(_ServiceAccount):
    __doc__ = _ServiceAccount.__doc__
    _asyncio = False


@sync
class ControllerRevision(_ControllerRevision):
    __doc__ = _ControllerRevision.__doc__
    _asyncio = False


@sync
class DaemonSet(_DaemonSet):
    __doc__ = _DaemonSet.__doc__
    _asyncio = False


@sync
class Deployment(_Deployment):
    __doc__ = _Deployment.__doc__
    _asyncio = False


@sync
class ReplicaSet(_ReplicaSet):
    __doc__ = _ReplicaSet.__doc__
    _asyncio = False


@sync
class StatefulSet(_StatefulSet):
    __doc__ = _StatefulSet.__doc__
    _asyncio = False


@sync
class HorizontalPodAutoscaler(_HorizontalPodAutoscaler):
    __doc__ = _HorizontalPodAutoscaler.__doc__
    _asyncio = False


@sync
class CronJob(_CronJob):
    __doc__ = _CronJob.__doc__
    _asyncio = False


@sync
class Job(_Job):
    __doc__ = _Job.__doc__
    _asyncio = False


@sync
class Ingress(_Ingress):
    __doc__ = _Ingress.__doc__
    _asyncio = False


@sync
class IngressClass(_IngressClass):
    __doc__ = _IngressClass.__doc__
    _asyncio = False


@sync
class NetworkPolicy(_NetworkPolicy):
    __doc__ = _NetworkPolicy.__doc__
    _asyncio = False


@sync
class PodDisruptionBudget(_PodDisruptionBudget):
    __doc__ = _PodDisruptionBudget.__doc__
    _asyncio = False


@sync
class ClusterRoleBinding(_ClusterRoleBinding):
    __doc__ = _ClusterRoleBinding.__doc__
    _asyncio = False


@sync
class ClusterRole(_ClusterRole):
    __doc__ = _ClusterRole.__doc__
    _asyncio = False


@sync
class RoleBinding(_RoleBinding):
    __doc__ = _RoleBinding.__doc__
    _asyncio = False


@sync
class Role(_Role):
    __doc__ = _Role.__doc__
    _asyncio = False


@sync
class CustomResourceDefinition(_CustomResourceDefinition):
    __doc__ = _CustomResourceDefinition.__doc__
    _asyncio = False


@sync
class Table(_Table):
    __doc__ = _Table.__doc__
    _asyncio = False


object_from_name_type = run_sync(partial(_object_from_name_type, _asyncio=False))
objects_from_files = run_sync(partial(_objects_from_files, _asyncio=False))
get_class = partial(_get_class, _asyncio=False)
new_class = partial(_new_class, asyncio=False)
object_from_spec = partial(_object_from_spec, _asyncio=False)



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import asyncio
import threading
from typing import Optional

from kr8s._api import Api as _AsyncApi
from kr8s._api import hash_kwargs


async def api(
    url: Optional[str] = None,
    kubeconfig: Optional[str] = None,
    serviceaccount: Optional[str] = None,
    namespace: Optional[str] = None,
    context: Optional[str] = None,
    _asyncio: bool = True,
) -> _AsyncApi:
    """Create a `kr8s.asyncio.Api` object for interacting with the Kubernetes API.

    If a kr8s object already exists with the same arguments in this thread, it will be returned.

    Args:
        url: The URL of the Kubernetes API server
        kubeconfig: The path to a kubeconfig file to use
        serviceaccount: The path of a service account to use
        namespace: The namespace to use
        context: The context to use

    Returns:
        kr8s.asyncio.Api: The API object

    Examples:
        >>> import kr8s
        >>> api = await kr8s.asyncio.api()  # Uses the default kubeconfig
        >>> print(await api.version())  # Get the Kubernetes version
    """
    from kr8s import Api as _SyncApi

    if _asyncio:
        _cls = _AsyncApi
    else:
        _cls = _SyncApi

    async def _f(**kwargs):
        key = hash_kwargs(kwargs)
        thread_id = threading.get_ident()
        try:
            loop_id = id(asyncio.get_running_loop())
        except RuntimeError:
            loop_id = 0
        thread_loop_id = f"{thread_id}.{loop_id}"
        if (
            _cls._instances
            and thread_loop_id in _cls._instances
            and key in _cls._instances[thread_loop_id]
        ):
            return await _cls._instances[thread_loop_id][key]
        if (
            all(k is None for k in kwargs.values())
            and thread_loop_id in _cls._instances
            and list(_cls._instances[thread_loop_id].values())
        ):
            return await list(_cls._instances[thread_loop_id].values())[0]
        return await _cls(**kwargs, bypass_factory=True)

    return await _f(
        url=url,
        kubeconfig=kubeconfig,
        serviceaccount=serviceaccount,
        namespace=namespace,
        context=context,
    )



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from typing import Dict, Optional, Type, Union

from kr8s._api import Api
from kr8s._objects import APIObject

from ._api import api as _api


async def get(
    kind: str,
    *names: str,
    namespace: Optional[str] = None,
    label_selector: Optional[Union[str, Dict]] = None,
    field_selector: Optional[Union[str, Dict]] = None,
    as_object: Optional[Type[APIObject]] = None,
    allow_unknown_type: bool = True,
    api=None,
    _asyncio=True,
    **kwargs,
):
    """Get a resource by name.

    This function retrieves a resource from the Kubernetes cluster based on its kind and name(s).
    It supports various options for filtering and customization.

    Args:
        kind: The kind of resource to get.
        *names: The names of the resources to get.
        namespace: The namespace to get the resource from.
        label_selector: The label selector to filter the resources by.
        field_selector: The field selector to filter the resources by.
        as_object: The object to populate with the resource data.
        allow_unknown_type: Automatically create a class for the resource if none exists.
        api: The api to use to get the resource.
        **kwargs: Additional keyword arguments to pass to the `httpx` API call.

    Returns:
        List[APIObject]: The Kubernetes resource objects.

    Raises:
        ValueError: If the resource is not found.

    Examples:
        >>> import kr8s
        >>> # All of these are equivalent
        >>> ings = await kr8s.asyncio.get("ing")                           # Short name
        >>> ings = await kr8s.asyncio.get("ingress")                       # Singular
        >>> ings = await kr8s.asyncio.get("ingresses")                     # Plural
        >>> ings = await kr8s.asyncio.get("Ingress")                       # Title
        >>> ings = await kr8s.asyncio.get("ingress.networking.k8s.io")     # Full group name
        >>> ings = await kr8s.asyncio.get("ingress.v1.networking.k8s.io")  # Full with explicit version
        >>> ings = await kr8s.asyncio.get("ingress.networking.k8s.io/v1")  # Full with explicit version alt.
    """
    if api is None:
        api = await _api(_asyncio=_asyncio)
    return await api.async_get(
        kind,
        *names,
        namespace=namespace,
        label_selector=label_selector,
        field_selector=field_selector,
        as_object=as_object,
        allow_unknown_type=allow_unknown_type,
        **kwargs,
    )


async def version(api=None, _asyncio=True):
    if api is None:
        api = await _api(_asyncio=_asyncio)
    return await api.async_version()


async def watch(
    kind: str,
    namespace: Optional[str] = None,
    label_selector: Optional[Union[str, Dict]] = None,
    field_selector: Optional[Union[str, Dict]] = None,
    since: Optional[str] = None,
    api=None,
    _asyncio=True,
):
    if api is None:
        api = await _api(_asyncio=_asyncio)
    async for t, o in api.async_watch(
        kind=kind,
        namespace=namespace,
        label_selector=label_selector,
        field_selector=field_selector,
        since=since,
    ):
        yield (t, o)


async def api_resources(api=None, _asyncio=True):
    if api is None:
        api = await _api(_asyncio=_asyncio)
    return await api.async_api_resources()


async def whoami(api=None, _asyncio=True):
    if api is None:
        api = await _api(_asyncio=_asyncio)
    return await api.async_whoami()


get.__doc__ = Api.get.__doc__
version.__doc__ = Api.version.__doc__
watch.__doc__ = Api.watch.__doc__
api_resources.__doc__ = Api.api_resources.__doc__
whoami.__doc__ = Api.whoami.__doc__



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from rich.console import Console

import kr8s

console = Console()


async def api_versions():
    """Print the supported API versions on the server, in the form of "group/version".

    Examples:
        # Print the supported API versions
        kubectl-ng api-versions

    """
    api = kr8s.api()
    for version in sorted(api.api_versions()):
        console.print(version)



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""Objects for managing a port forward connection.

This module provides a class for managing a port forward connection to a Kubernetes Pod or Service.
"""
from kr8s._portforward import PortForward

__all__ = ["PortForward"]



# SPDX-FileCopyrightText: Copyright (c) 2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import pathlib

from kubectl_ng.cli import app
from typer.testing import CliRunner

from kr8s.objects import objects_from_files

runner = CliRunner()

HERE = pathlib.Path(__file__).parent.absolute()


def test_create_and_delete():
    spec = str(HERE / "resources" / "simple" / "nginx_pod_service.yaml")

    objs = objects_from_files(spec)
    for obj in objs:
        assert not obj.exists()

    result = runner.invoke(app, ["create", "-f", spec])
    assert result.exit_code == 0
    for obj in objs:
        assert obj.name in result.stdout

    for obj in objs:
        assert obj.exists()

    result = runner.invoke(app, ["delete", "-f", spec])
    assert result.exit_code == 0
    for obj in objs:
        assert obj.name in result.stdout

    for obj in objs:
        assert not obj.exists()



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from __future__ import annotations

import json
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncGenerator, BinaryIO

from kr8s._exceptions import ExecError

if TYPE_CHECKING:
    from kr8s._objects import APIObject

STDIN_CHANNEL: int = 0
STDOUT_CHANNEL: int = 1
STDERR_CHANNEL: int = 2
ERROR_CHANNEL: int = 3
RESIZE_CHANNEL: int = 4
CLOSE_CHANNEL: int = 255
EXEC_PROTOCOL: str = "v4.channel.k8s.io"


class Exec:
    """Executes a command in a running container."""

    def __init__(
        self,
        resource: APIObject,
        command: list[str],
        container: str | None = None,
        stdin: str | BinaryIO | None = None,
        stdout: BinaryIO | None = None,
        stderr: BinaryIO | None = None,
        check: bool = True,
        capture_output: bool = True,
    ) -> None:
        self._resource = resource
        self._container = container

        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr
        self._capture_output = capture_output

        self.args = command
        self.stdout = b""
        self.stderr = b""
        self.returncode: int
        self.check = check

    @asynccontextmanager
    async def run(
        self,
    ) -> AsyncGenerator[Exec, CompletedExec]:
        assert self._resource.api
        async with self._resource.api.open_websocket(
            version=self._resource.version,
            url=f"{self._resource.endpoint}/{self._resource.name}/exec",
            subprotocols=(EXEC_PROTOCOL,),
            namespace=self._resource.namespace,
            params={
                "command": self.args,
                "container": self._container or self._resource.spec.containers[0].name,
                "stdout": "true" if self._stdout or self._capture_output else "false",
                "stderr": "true" if self._stderr or self._capture_output else "false",
                "stdin": "true" if self._stdin is not None else "false",
            },
        ) as ws:
            if self._stdin:
                if ws.subprotocol != "v5.channel.k8s.io":
                    raise ExecError(
                        "Stdin is not supported with protocol "
                        f"{ws.subprotocol}, only with v5.channel.k8s.io"
                    )
                if isinstance(self._stdin, str):
                    await ws.send_bytes(STDIN_CHANNEL.to_bytes() + self._stdin.encode())  # type: ignore
                else:
                    await ws.send_bytes(STDIN_CHANNEL.to_bytes() + self._stdin.read())  # type: ignore
                await ws.send_bytes(CLOSE_CHANNEL.to_bytes() + STDIN_CHANNEL.to_bytes())  # type: ignore
            while True:
                message = await ws.receive_bytes()
                channel, message = int(message[0]), message[1:]
                if message:
                    if channel == STDOUT_CHANNEL:
                        if self._capture_output:
                            self.stdout += message
                        if self._stdout:
                            self._stdout.write(message)
                    elif channel == STDERR_CHANNEL:
                        if self._capture_output:
                            self.stderr += message
                        if self._stderr:
                            self._stderr.write(message)
                    elif channel == ERROR_CHANNEL:
                        error = json.loads(message.decode())
                        if error["status"] == "Success":
                            self.returncode = 0
                            break
                        # Extract return code from details
                        if "details" in error and "causes" in error["details"]:
                            for cause in error["details"]["causes"]:
                                if "reason" in cause and cause["reason"] == "ExitCode":
                                    self.returncode = int(cause["message"])
                                    break
                        else:
                            self.returncode = 1
                        if self.check:
                            raise ExecError(error["message"])
                        break
                    else:
                        raise ExecError(
                            f"Unhandled message on channel {channel}: {message}"
                        )
            yield self

    async def wait(self) -> CompletedExec:
        return self.as_completed()

    def as_completed(self) -> CompletedExec:
        return CompletedExec(
            args=self.args,
            stdout=self.stdout,
            stderr=self.stderr,
            returncode=self.returncode,
        )


@dataclass
class CompletedExec:
    """Result of an Exec command.

    Similar to subprocess.CompletedProcess.
    """

    args: str | list[str]
    stdout: bytes
    stderr: bytes
    returncode: int

    def check_returncode(self) -> None:
        if self.returncode != 0:
            raise ExecError(f"Command {self.args} exited with status {self.returncode}")



# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import contextlib
import os
from typing import Generator


@contextlib.contextmanager
def set_env(**environ: str) -> Generator[None, None, None]:
    """Temporarily sets the process environment variables.

    This context manager allows you to temporarily set the process environment variables
    within a specific scope. It saves the current environment variables, updates them with
    the provided values, and restores the original environment variables when the scope
    is exited.

    Args:
        **environ: Keyword arguments representing the environment variables to set.

    Yields:
        None

    Examples:
        >>> with set_env(PLUGINS_DIR='test/plugins'):
        ...     "PLUGINS_DIR" in os.environ
        True

        >>> "PLUGINS_DIR" in os.environ
        False

    """
    old_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)
